{******************************************************************************}
{                                                                              }
{       ga SQL Parser package                                                  }
{       gaDoubleLinklist classes for internal data storage                     }
{                                                                              }
{       Copyright (c) 2001, 2002 AS Gaiasoft                                   }
{       Portions copyright (c) Julian M Bucknall, from Algorithms Alfresco in  }
{       The Delphi Magazine (www.thedelphimagazine.com)                        }
{                                                                              }
{       Redesigned by Gert Kello                                               }
{                                                                              }
{******************************************************************************}

//: Unit for gaLinkList classes 
{:
Unit for gaLinkList and gaListBookmark classes
}
unit gaLnkList;

interface

uses
  SysUtils, Classes;

type
  TgaNodeNotification = (nnAdded, nnDeleted, nnCleared);

type
  PdllNode = ^TdllNode;
  TdllNode = packed record
    dllnNext : PdllNode;
    dllnPrev : PdllNode;
    dllnData : Pointer;
  end;

  //: Abstract list bookmark class 
  {:
  Abstract base class for list bookmarks.
  Contains properties to determ wheter the bookmark is valid or not and to get
  read/write access to the item pointed by bookmark.
  Descendants of the class should at least override the access methods
  GetIsValid, GetItem and SetItem.
  They would propably need to add property to store the bookmark pointer and
  the list for which the bookmark is created.
  }
  TgaListBookmark = class (TObject)
  protected
    //: GetIsValid is the abstract read access method of the IsValid property. 
    function GetIsValid: Boolean; virtual; abstract;
    //: GetItem is the read access method of the Item property 
    function GetItem: Pointer; virtual; abstract;
    //: SetItem is the write access method of the Item property. 
    procedure SetItem(Value: Pointer); virtual; abstract;
  public
    //: Property IsValid determs whether the bookmark is valid or not 
    {:
    Property IsValid is read and run time only.
    It determs whether the bookmark is valid or not.
    }
    property IsValid: Boolean read GetIsValid;
    //: Property Item gives an read/write access to the item represented by the bookmark 
    {:
    Property Item is read / write, at run time only.
    It gives an read/write access to the item represented by the cursor.
    The descendants are required to implement the access methods GetItem and
    SetItem.
    }
    property Item: Pointer read GetItem write SetItem;
  end;
  
  TgaSimpleDoubleList = class (TObject)
  private
    //: FHead is the state field of the Head property. 
    {:
    FHead is the state field of the Head property.
    }
    FHead: PdllNode;
    {:
    FTail is the state field of the Tail property.
    }
    FTail: PdllNode;
    //: GetHead is the read access method of the Head property. 
    function GetHead: PdllNode;
    //: GetTail is the read access method of the Tail property. 
    function GetTail: PdllNode;
  protected
    //: Checks and corrects the Head and Tail node of the list 
    procedure CheckListBorders; virtual;
    procedure InitDoubleList; virtual;
    procedure InternalDelete(ANode: PdllNode);
    function InternalInsert(AItem : pointer; AInsertItem: PdllNode): PdllNode;
    class function IsList(AHead, ATail: PdllNode): Boolean;
    function IsNodeInside(ANode: PdllNode): Boolean;
    function NodeOf(AItem: pointer): PdllNode;
    procedure Notify(Ptr: Pointer; Action: TListNotification); virtual;
    procedure NotifyChange(AOldItem, ANewItem: pointer);
    procedure NotifyNodeChange(ANode: PdllNode; Action: TgaNodeNotification); 
            virtual;
    {:
    Property Head is read only.
    }
    property Head: PdllNode read GetHead;
    {:
    Property Tail is read only.
    It represents the node after the last node in the list.
    If List has a cursor then Eof := Cursor = Tail;
    }
    property Tail: PdllNode read GetTail;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear; virtual;
  end;
  
  TgaSharedDoubleList = class;

  TgaDoubleList = class (TgaSimpleDoubleList)
  private
    {:
    Field FActiveDataUserList.
    }
    FActiveDataUserList: TgaSimpleDoubleList;
    {:
    Field FDataUserList.
    }
    FDataUserList: TgaSimpleDoubleList;
    {:
    FCount is the state field of the Count property.
    }
    FCount: Integer;
    {:
    FCursor is the state field of the Cursor property.
    }
    FCursor: PdllNode;
    FNodeRevID: Cardinal;
    function GetBof: Boolean;
    function GetCursor: PdllNode;
    function GetEof: Boolean;
    function GetFirstItem: Pointer;
    function GetLastItem: Pointer;
    procedure SetFirstItem(Value: Pointer);
    procedure SetLastItem(Value: Pointer);
  protected
    procedure FreeDataUserList;
    procedure AddActiveDataUser(ADataUser: TgaSharedDoubleList); virtual;
    procedure AddDataUser(ADataUser: TgaSharedDoubleList); virtual;
    procedure RemoveActiveDataUser(ADataUser: TgaSharedDoubleList); virtual;
    procedure RemoveDataUser(ADataUser: TgaSharedDoubleList); virtual;
    procedure CheckListCursor; virtual;
    function Extract(Item: Pointer): Pointer;
    function GetCurrentItem: Pointer; virtual;
    function GetIsEmpty: Boolean; virtual;
    procedure InitDoubleList; override;
    function IsCursorCorrect: Boolean; virtual;
    procedure NotifyNodeChange(ANode: PdllNode; Action: TgaNodeNotification); 
            override;
    procedure SetCurrentItem(Value: Pointer); virtual;
    {:
    Property Cursor is read only.
    }
    property Cursor: PdllNode read GetCursor;
    {:
    A number wich is icremented for every list node change (for dataowner lists)
    or dtores the last state when the list was validated against the DataOwner 
    list
    }
    property NodeRevID: Cardinal read FNodeRevID;
  public
    destructor Destroy; override;
    procedure Add(Item: Pointer); virtual;
    procedure CopyListContest(ACopyList: TgaDoubleList); virtual;
    procedure DeleteCurrent;
    procedure First;
    function GetBookmark: TgaListBookmark; virtual;
    procedure GotoBookmark(ABookmark: TgaListBookmark);
    procedure InsertAfterCurrent(AItem: Pointer; MoveCursorToNewItem: boolean); 
            virtual;
    procedure Last;
    function Locate(AItem: Pointer): Boolean;
    procedure Next;
    procedure Pack;
    procedure Previous;
    procedure Remove(Item: Pointer);
    {:
    Property Bof is read and run time only.
    }
    property Bof: Boolean read GetBof;
    {:
    Property Count is read and run time only.
    }
    property Count: Integer read FCount;
    {:
    Property CurrentItem is read / write, at run time only.
    }
    property CurrentItem: Pointer read GetCurrentItem write SetCurrentItem;
    {:
    Property Eof is read and run time only.
    }
    property Eof: Boolean read GetEof;
    {:
    Property FirstItem is read / write, at run time only.
    }
    property FirstItem: Pointer read GetFirstItem write SetFirstItem;
    {:
    Property IsEmpty is read and run time only.
    }
    property IsEmpty: Boolean read GetIsEmpty;
    {:
    Property LastItem is read / write, at run time only.
    }
    property LastItem: Pointer read GetLastItem write SetLastItem;
  end;
  
  TgaSharedDoubleList = class (TgaDoubleList)
  private
    {:
    FDataOwner is the state field of the DataOwner property.
    }
    FDataOwner: TgaDoubleList;
    {:
    Field FFirstNode.
    }
    FFirstNode: PdllNode;
    {:
    Field FLastNode.
    }
    FLastNode: PdllNode;
    {:
    FActiveDataShare is the state field of the ActiveDataShare property.
    }
    FActiveDataShare: Boolean;
    FCursorNodeRev: Cardinal;
    {:
    FIsDataOwner is the state field of the IsDataOwner property.
    }
    FIsDataOwner: Boolean;
    procedure SetActiveDataShare(Value: Boolean);
  protected
    procedure DataOwnerDestroyed;
    procedure InitDoubleList; override;
    procedure AddActiveDataUser(ADataUser: TgaSharedDoubleList); override;
    procedure AddDataUser(ADataUser: TgaSharedDoubleList); override;
    procedure CheckListBorders; override;
    procedure CheckListCursor; override;
    //: Calls Dataowners Notify in addition if the list is mirroring list 
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
    //: Called from DataOwner.NotifyNodeChange if ActiveDataShare = True 
    procedure NotifyMasterChanged(ANode: PdllNode; Action: TgaNodeNotification);
            virtual;
    //: Calls inherited if IsDataOwner, otherwise calls DataOwners NotifyNodeChange 
    procedure NotifyNodeChange(ANode: PdllNode; Action: TgaNodeNotification); 
            override;
    procedure RemoveActiveDataUser(ADataUser: TgaSharedDoubleList); override;
    procedure RemoveDataUser(ADataUser: TgaSharedDoubleList); override;
    procedure ValidateCursor(CorrectIfInvalid: Boolean); virtual;
    procedure ValidateListBorders; virtual;
  public
    constructor Create;
    constructor CreateMirror(AMirroredList: TgaDoubleList);
    destructor Destroy; override;
    procedure SetEndPos(APositionedList: TgaDoubleList; StrictEndPos: Boolean);
    procedure SetStartPos(APositionedList: TgaDoubleList; StrictStartPos: 
            boolean);
    procedure Add(Item: Pointer); override;
    procedure Clear; override;
    //: procedure InsertAfterCurrent overrides inherited InsertAfterCurrent. 
    procedure InsertAfterCurrent(AItem: Pointer; MoveCursorToNewItem: boolean); 
            override;
    {:
    Property DataOwner is read and run time only.
    if DataOwner = Self, then IsDataOwner = True;
    the IsDataOwner and IsDataOwner proeprties are set in the constructor and
    can't be changed during the life of the list
    }
    property DataOwner: TgaDoubleList read FDataOwner;
    {:
    Property ActiveDataShare is read / write, at run time only.
    Determs wheter the mirroring list is notified by the master list with the
    call to NotifyMasterChanged
    }
    property ActiveDataShare: Boolean read FActiveDataShare write 
            SetActiveDataShare;
    {:
    Property IsDataOwner is read and run time only.
    }
    property IsDataOwner: Boolean read FIsDataOwner;
  end;
  
  //: Bookmark for the TgaDoubleList, TgaSharedDoubleList and their descendants 
  {:
  TgaDoubleListBookmark implements abstract methods (GetIsValid, GetItem and
  SetItem) of the TgaListBookmark. It implements functionality to work as
  Bookmark for TgaDoubleLinkList, TgaSharedDoubleLinkList and their descendants.
  }
  TgaDoubleListBookmark = class (TgaListBookmark)
  private
    {:
    FCursor is the state field of the Cursor property.
    }
    FCursor: PdllNode;
    FNodeRevID: Cardinal;
    {:
    FOwnerList is the state field of the OwnerList property.
    }
    FOwnerList: TgaDoubleList;
  protected
    //: Overrides abstarct TgaListBookmark.GetIsValid 
    function GetIsValid: Boolean; override;
    //: Function GetItem overrides abstract GetItem 
    function GetItem: Pointer; override;
    function IsValidFor(AList: TgaDoubleList): Boolean; virtual;
    procedure SetItem(Value: Pointer); override;
    {:
    Property Cursor is read only.
    }
    property Cursor: PdllNode read FCursor;
    property NodeRevID: Cardinal read FNodeRevID;
    {:
    Property OwnerList is read only.
    }
    property OwnerList: TgaDoubleList read FOwnerList;
  public
    constructor Create(AOwnerList: TgaDoubleList);
  end;
  
  EgaLinkListError = class (Exception)
  end;
  
  EgaListBookmarkError = class (Exception)
  end;
  
implementation

resourcestring
// START resource string wizard section
  SListIsNotAMirroringList = 'List is not a mirroring list: %s';
  SListIsNotDataOwner = 'List is not the data owner: %s';
  SListIsDataOwner = 'List is the data owner: %s';
  SListsDoNotShareData = 'Lists do not share data: %s';
  SMirrorListValidateFailed = 'Mirror list validate failed: %s';
  SInvalidBookmark = 'Invalid bookmark: %s';
  SCantSetTheStartOrEndPostions = 'Can''t set the start or end postions';
  SStrictFirstNodeNotPartOfTheMasterList = 'Strict first node is invalid - not part of the master list';
  SStrictFirstNodePointsToTheHeadOfMasterList = 'Strict first node is invalid - points to the Head item of master list';
  SCantGetItem = 'Can''t get the item';
  SCantSetItem = 'Can''t set the item';
  SBookmarkClassInvalid = 'Bookmark is a member of invalid class';
  SHeadNodeNotInMasterList = 'Head node is invalid - not part of the master list';
  SInvalidCursor = 'Cursor is not part of the list';
  SNoHeadNode = 'Invalid link list - no head node';
  SCantAddActiveDataUser = 'Can''t add active data user';
  SCantAddDataUser = 'Can''t add data user';
  SCantRemoveActiveDataUser = 'Can''t remove active data user';
  SCantRemoveDataUser = 'Can''t remove data user';
  SNoCurrentItemForTheOperation = 'Either Bof or Eof is true - no current item for the operation';
  SCantChangeActiveDataShare = 'Can''t change ActiveDataShare';
  SCannotDeleteHeadOrTailNode = 'Can''t delete Head or Tail node';
  SNoItemsInTheList = 'There is no items in the list';
  SInvalidHeadOrTail = 'Invalid Head or Tail node - possibly Head after Tail or Tail not a part of master list';

// END resource string wizard section

procedure dnmFreeNode(aNode : PdllNode);
begin
  Dispose(aNode);
end;

function dnmAllocNode : PdllNode;
begin
  New(Result);
end;

{:
GetIsValid is the read access method of the IsValid property.
Descendant classes should override this method to give appropriate result
}
{:
GetItem is the read access method of the Item property.
It retrives the value of the item represented by the cursor. The method has 
to be overridden by the descendant classes.
}
{:
SetItem is the write access method of the Item property.
It changes the value of the item represented by the cursor. The method has 
to be overridden by the descendant classes.
}
{
***************************** TgaSimpleDoubleList ******************************
}
constructor TgaSimpleDoubleList.Create;
{:
Constructor Create overrides the inherited Create.
First inherited Create is called, then the internal data structure is
initialized via call to InitDoubleList;
}
begin
  inherited Create;
  InitDoubleList;
end;

destructor TgaSimpleDoubleList.Destroy;
{:
Destructor Destroy overrides the inherited Destroy.
First the Clear methodis called, then all owned fields are free'd,
finally inherited Destroy is called.
}
begin
  Clear;
  if Assigned(FHead) then
    dnmFreeNode(FHead);
  if Assigned(FTail) then
    dnmFreeNode(FTail);
  inherited Destroy;
end;

procedure TgaSimpleDoubleList.CheckListBorders;
{:
procedure CheckListBorders - Checks that the Head and Tail node are correct.
Corrects them if required.
This method is called only from GetTail and GetHead methods.
It should be as optimized as much as possible because it get's called every
time the Head or Tail node is used (getEof or GetBof for example).
For TgaSimpleDoubleList and TgaDoubleList, theese nodes are always correct.
}
begin
  ; // TgaSimpleDoubleList and TgaDoubleList have always correct border nodes
end;

procedure TgaSimpleDoubleList.Clear;
{:
procedure Clear.
Clears the contest of the list.
It first calls NotifyNodeChange with nil ChangedNode (to indicate that zero
or more nodes might be affected) and nnCleared action. Then, for every
item in the list, the holding node is removed and Notify is called to indicate
that the item has been deleted.
This form of clear is not allowed for mirroring lists - they must not call
NotifyNodeChange with nnCleared flag.
}
var
  Temp: PdllNode;
  tmpData: Pointer;
begin
  if Assigned(FHead) then
  begin
    { There is no need to correct list state here - this form of clear can be
      executed *ONLY* if the list is master list }
    NotifyNodeChange(nil, nnCleared);
    Temp := FHead^.dllnNext;
    while (Temp <> FTail) do begin
      FHead^.dllnNext := Temp^.dllnNext;
      tmpData := Temp^.dllnData;
      dnmFreeNode(Temp);
      if Assigned(tmpData) then
        Notify(tmpData, lnDeleted);
      Temp := FHead^.dllnNext;
    end;
    FTail^.dllnPrev := FHead;
  end;
end;

function TgaSimpleDoubleList.GetHead: PdllNode;
{:
GetHead is the read access method of the Head property.
It checks (and corrects if needed) the head and tail node of the list trhough
the call to CheckListBorders.
}
begin
  CheckListBorders;
  Result := FHead;
end;

function TgaSimpleDoubleList.GetTail: PdllNode;
{:
GetTail is the read access method of the Tail property.
It checks (and corrects if needed) the head and tail node of the list trhough
the call to CheckListBorders.
}
begin
  CheckListBorders;
  Result := FTail;
end;

procedure TgaSimpleDoubleList.InitDoubleList;
{:
procedure InitDoubleList.
Allocates Head and Tail node and initializes them to form a list
}
begin
  {allocate a head and a tail node}
  FHead := dnmAllocNode;
  FTail := dnmAllocNode;
  FHead^.dllnNext := FTail;
  FHead^.dllnPrev := nil;
  FHead^.dllnData := nil;
  FTail^.dllnNext := nil;
  FTail^.dllnPrev := FHead;
  FTail^.dllnData := nil;
end;

procedure TgaSimpleDoubleList.InternalDelete(ANode: PdllNode);
{:
procedure InternalDelete.
ANode - A node to be deleted.
First the node is removed from list.
Then the list is notified about the node change (call to NotifyNodeChange)
After the that node is disposed and Notify is called for Nodes Data (item)
pointer
All list deletions should be done trough this method (the Clear is exeception).
}
var
  tmpData: Pointer;
begin
  if (ANode = FHead) or (ANode = FTail) then
    raise EgaLinkListError.Create(SCannotDeleteHeadOrTailNode);
  tmpData := ANode^.dllnData;
  ANode^.dllnPrev^.dllnNext := ANode^.dllnNext;
  ANode^.dllnNext^.dllnPrev := ANode^.dllnPrev;
  NotifyNodeChange(ANode, nnDeleted);
  dnmFreeNode(ANode);
  if Assigned(tmpData) then
    Notify(tmpData, lnDeleted);
end;

function TgaSimpleDoubleList.InternalInsert(AItem : pointer; AInsertItem: 
        PdllNode): PdllNode;
{:
function InternalInsert.
AItem - A pointer to item to be inserted to the list,
AInsertItem - A item after which the new item is inserted.
Returns: A node which holds the e new item.
All list additions should be done trough this method.
}
begin
  if (AInsertItem = Tail) then
    AInsertItem := AInsertItem^.dllnPrev;
  {allocate a new node and insert after the AInsertItem}
  Result := dnmAllocNode;
  Result^.dllnData := aItem;
  Result^.dllnNext := AInsertItem^.dllnNext;
  Result^.dllnPrev := AInsertItem;
  AInsertItem^.dllnNext := Result;
  Result^.dllnNext^.dllnPrev := Result;
  NotifyNodeChange(Result, nnAdded);
  if Assigned(AItem) then
    Notify(AItem, lnAdded);
end;

class function TgaSimpleDoubleList.IsList(AHead, ATail: PdllNode): Boolean;
{:
class function IsList - checks whether two nodes are connected.
AHead, ATail - nodes to be tested for connectivity.
Returns: True, if it is possible to walk from the AHead node to the ATail node.
}
var
  tmpNode: PdllNode;
begin
  tmpNode := AHead;
  while Assigned(tmpNode) and (tmpNode <> ATail) do
    tmpNode := tmpNode^.dllnNext;
  Result := tmpNode = ATail;
end;

function TgaSimpleDoubleList.IsNodeInside(ANode: PdllNode): Boolean;
{:
function IsNodeInside - checks whether the node is inside the list.
ANode - a node to be tested.
Returns: True, if the Node is between the Head and Tail
nodes (inclusive). Head and Tail nodes are considered to be inside the list
}
var
  tmpCursor: PdllNode;
  tmpTail: PdllNode;
begin
  tmpCursor := Head;
  tmpTail := Tail;
  if Assigned(tmpCursor) and Assigned(ANode) then
  begin
    while (tmpCursor <> ANode) and (tmpCursor <> tmpTail) do
      tmpCursor := tmpCursor^.dllnNext;
    Result := tmpCursor = ANode;
  end else
    Result := False;
end;

function TgaSimpleDoubleList.NodeOf(AItem: pointer): PdllNode;
{:
function NodeOf - finds a first node that holds a item pointer.
AItem - pointer which is searched for.
Returns: pointer to the first node, that holds the AItem.
}
var
  tmpTail: PdllNode;
begin
  if Head = nil then
    raise EgaLinkListError.Create(SNoHeadNode);
  Result := Head.dllnNext;
  tmpTail := Tail;
  while (Result <> tmpTail) and (Result^.dllnData <> AItem) do
    Result := Result^.dllnNext;
  if Result = tmpTail then
    Result := nil;
end;

procedure TgaSimpleDoubleList.Notify(Ptr: Pointer; Action: TListNotification);
{:
procedure Notify. Notify is called each time the logical contest of list
changes. Logical Conetnts means the set of items the form the list. In example,
adding or removing nil item is not considered to be an change in logical
conents, but replacing a item with another one is.
Ptr - pointer to the item for which the Action is performed.
}
begin
  ;// Do nothing here
end;

procedure TgaSimpleDoubleList.NotifyChange(AOldItem, ANewItem: pointer);
{:
NotifySwap calls lnDeleted notify for AOldItem, if assigned, and lnAdded notify 
for ANewItem, if Assigned.
}
begin
  if AOldItem <> ANewItem then
  begin
    if Assigned(AOldItem) then
      Notify(AOldItem, lnDeleted);
    if Assigned(ANewItem) then
      Notify(ANewItem, lnAdded);
  end;
end;

procedure TgaSimpleDoubleList.NotifyNodeChange(ANode: PdllNode; Action: 
        TgaNodeNotification);
{:
procedure NotifyNodeChange - A method that is meant for checking internal data
dependencies, for example whtere the node being removed is cursor or not.
ANode - the node that is affected by the Action.
Tere is no dependecies to check in TgaSimpleDoubleList.
}
begin
  ;// Do nothing here
end;

{
******************************** TgaDoubleList *********************************
}
destructor TgaDoubleList.Destroy;
{:
Destructor Destroy overrides the inherited Destroy.
First the Cursor is cleared (to avoid problems in Clear), then the DataUserList
checked for and freed if present, finally inherited Destroy is called.
}
begin
  FCursor := nil;
  { FreeDataUserList will also free FActiveDataUserList. And if the DataUserList
    is unassigned, then the ActiveDataUserList should also be.. }
  if Assigned(FDataUserList) then
    FreeDataUserList;
  inherited Destroy;
end;

procedure TgaDoubleList.Add(Item: Pointer);
{:
procedure Add - Adds a item at the end of the list.
Item - item to be added.
}
begin
  InternalInsert(Item, Tail);
end;

procedure TgaDoubleList.AddActiveDataUser(ADataUser: TgaSharedDoubleList);
{:
procedure AddActiveDataUser - Adds a mirroring list (ADataUser) to the internal
list. This list is used to notify DataUsers about the Node changes inside the
master list.
}
begin
  if not Assigned(FActiveDataUserList) then
    FActiveDataUserList := TgaSimpleDoubleList.Create;
  FActiveDataUserList.InternalInsert(ADataUser, FActiveDataUserList.Tail);
end;

procedure TgaDoubleList.AddDataUser(ADataUser: TgaSharedDoubleList);
{:
procedure AddDataUser - Adds a list to the internal DataUsers list, so that it
can be notified if the DataOwner is destroyed. Called (normally) from the
TgaSharedDoubleLinkList.CreateMirror.
}
begin
  if not Assigned(FDataUserList) then
    FDataUserList := TgaSimpleDoubleList.Create;
  FDataUserList.InternalInsert(ADataUser, FDataUserList.FTail);
end;

procedure TgaDoubleList.CheckListCursor;
{:
procedure CheckListCursor - Checks the cursor of the list. Raises an Exception
if cursor is incorrect. For TgaDoubleList, the cursor is always correct.
This method should be normally called from property Cursor read access method.
Also, it might get called from other methods that deal with current record but
the cursor is refferred trough the FCursor field rather that Cursor property
}
begin
  ; // TgaSimpleDoubleList and TgaDoubleList are always in correct state
end;

procedure TgaDoubleList.CopyListContest(ACopyList: TgaDoubleList);
{:
procedure CopyListContest - Clears self and then adds all items from the
ACopyList to the self.
}
begin
  Clear;
  ACopyList.First;
  while not ACopyList.Eof do
  begin
    Add(ACopyList.CurrentItem);
    ACopyList.Next;
  end;
end;

procedure TgaDoubleList.DeleteCurrent;
{:
procedure DeleteCurrent - deletes the item represented by the cursor from the
list.
}
begin
  if Eof or Bof then
    raise EgaLinkListError.Create(SNoCurrentItemForTheOperation);
  InternalDelete(Cursor);
end;

function TgaDoubleList.Extract(Item: Pointer): Pointer;
{:
function Extract - removes the item from the list, but with lnExtracted
notification rather than lnDeleted notification.
Item - a item to be deleted.
Returns: Item, if the item was found from the list.
Nil, if the item was not found from the list.
}
var
  tmpNode: PdllNode;
begin
  Result := nil;
  tmpNode := NodeOf(Item);
  if Assigned(tmpNode) then
  begin
    Result := Item;
    tmpNode^.dllnData := nil;
    InternalDelete(tmpNode);
    Notify(Result, lnExtracted);
  end;
end;

procedure TgaDoubleList.First;
{:
procedure First - sets the cursor to point to the first item in the list.
If the list is empty, the item will be the Tail item. After the call to the
First method, the BOF is always False (but Eof might be true, in the case of
empty list).
}
begin
  FCursor := Head^.dllnNext;
end;

procedure TgaDoubleList.FreeDataUserList;
{:
procedure FreeDataUserList - Notifies all data users that the master list is
about to be destroyed. Clears and frees the FDataUserList and
FActiveDataUserList.
}
var
  tmpNode: PdllNode;
begin
  if Assigned(FDataUserList) then
  begin
    tmpNode := FDataUserList.FHead^.dllnNext;
    while tmpNode <> FDataUserList.FTail do
    begin
      TgaSharedDoubleList(tmpNode^.dllnData).DataOwnerDestroyed;
      tmpNode := tmpNode^.dllnNext;
    end;
    FreeAndNil(FDataUserList);
    FreeAndnIl(FActiveDataUserList);
  end;
end;

function TgaDoubleList.GetBof: Boolean;
{:
GetBof is the read access method of the Bof property.
Results true if the cursor is positioned at the head (before the first) node
of the list.
}
begin
  Result := Head = Cursor;
end;

function TgaDoubleList.GetBookmark: TgaListBookmark;
{:
function GetBookmark - generates the bookamark for the current item in the list.
Returns: Bookmark that was generated.
}
begin
  Result := TgaDoubleListBookmark.Create(Self);
end;

function TgaDoubleList.GetCurrentItem: Pointer;
{:
GetCurrentItem is the read access method of the CurrentItem property.
There are no current item if the list is either at the Head or Tail node.
}
begin
  if Eof or Bof then
    raise EgaLinkListError.Create(SNoCurrentItemForTheOperation);
  Result := Cursor.dllnData;
end;

function TgaDoubleList.GetCursor: PdllNode;
{:
GetCursor is the read access method of the Cursor property.
It calls the CheckListCursor method which raises an exception, if the cursor is
invalid at the given moment
}
begin
  CheckListCursor;
  Result := FCursor;
end;

function TgaDoubleList.GetEof: Boolean;
{:
GetEof is the read access method of the Eof property.
Results true if the cursor is positioned at the tail (after the last) node
of the list.
}
begin
  Result := Tail = Cursor;
end;

function TgaDoubleList.GetFirstItem: Pointer;
{:
GetFirstItem is the read access method of the FirstItem property.
It resturns the first item (the one after the Head node) in the list.
If the list is empty, the exeception is raised
}
begin
  if IsEmpty then
    raise EgaLinkListError.Create(SNoItemsInTheList);
  Result := Head^.dllnNext.dllnData;
end;

function TgaDoubleList.GetIsEmpty: Boolean;
{:
GetIsEmpty is the read access method of the IsEmpty property.
A list is empty, if the next node after the Head is the Tail node.
}
begin
  Result := Head^.dllnNext = Tail;
end;

function TgaDoubleList.GetLastItem: Pointer;
{:
GetLastItem is the read access method of the LastItem property.
It returns the last item (the one before the Tail node) in the list.
If the list is empty, the exeception is raised.
}
begin
  if IsEmpty then
    raise EgaLinkListError.Create(SNoItemsInTheList);
  Result := Tail^.dllnPrev.dllnData;
end;

procedure TgaDoubleList.GotoBookmark(ABookmark: TgaListBookmark);
{:
procedure GotoBookmark - postions the cuurent item of the list to the item
represented by the ABookmark. The ABookmark must be obtained from the same list
as GotoBookamrk is called, or the lists must share the data (they must have
common DataOwner).
The CheckListBorders is called to ensure that the list represents valid set of
data.
}
var
  tmpBookmark: TgaDoubleListBookmark;
begin
  CheckListBorders;
  if not (ABookmark is TgaDoubleListBookmark) then
    raise EgaLinkListError.Create(SBookmarkClassInvalid);
  tmpBookmark := ABookmark as TgaDoubleListBookmark;
  if not tmpBookmark.IsValidFor(Self) then
    raise EgaLinkListError.Create(SInvalidBookmark);
  FCursor := tmpBookMark.Cursor;
end;

procedure TgaDoubleList.InitDoubleList;
{:
procedure InitDoubleList overrides inherited InitDoubleList.
It first calls inherited InitDaoubleList, after that the Cursor, Count and
Node  Revision ID are initialised.
}
begin
  inherited InitDoubleList;
  { set the cursor to the head node and item count to the zero}
  FCursor := FHead;
  FCount := 0;
  FNodeRevID := 0;
end;

procedure TgaDoubleList.InsertAfterCurrent(AItem: Pointer; MoveCursorToNewItem: 
        boolean);
{:
procedure InsertAfterCurrent - Inserts a item represented by the AItem after the
cursor. If MoveCursorToNewItem = True, then the cursor is moved to the item
inserted.
}
begin
  if MoveCursorToNewItem then
    FCursor := InternalInsert(AItem, Cursor)
  else
    InternalInsert(AItem, Cursor);
end;

function TgaDoubleList.IsCursorCorrect: Boolean;
{:
function IsCursorCorrect - Returns a boolean indicating whether the node
pointed by the cursor is inside the list or .
}
begin
  Result := IsNodeInside(FCursor);
end;

procedure TgaDoubleList.Last;
{:
procedure Last - sets the cursor to point to the last item in the list.
If the list is empty, the item will be the Head item. After the call to the
last method, the EOF is always False (but Bof might be true, in the case of
empty list).
}
begin
  FCursor := Tail^.dllnPrev;
end;

function TgaDoubleList.Locate(AItem: Pointer): Boolean;
{:
function Locate -  postions cursor to the first node that contains AItem, if
one exists. If the node is found, it returns True. If not, it returns False.
}
var
  tmpCursor: PdllNode;
begin
  tmpCursor := NodeOf(AItem);
  if Assigned(tmpCursor) then
  begin
    FCursor := tmpCursor;
    Result := True;
  end else
    Result := False;
end;

procedure TgaDoubleList.Next;
{:
procedure Next moves the cursor to the next item on the list. If cursor is
already at the end of the list, it deos nothing.
}
begin
  if not Eof then
    FCursor := Cursor^.dllnNext;
end;

procedure TgaDoubleList.NotifyNodeChange(ANode: PdllNode; Action: 
        TgaNodeNotification);
{:
procedure NotifyNodeChange overrides inherited NotifyNodeChange calls inherited
NotifyNodeChange.
In addition, it corrects the item count of the list, changes cursor if required,
increments Node Revision ID and notifies Active mirroring lists about the
change.
}
var
  tmpCursor: PdllNode;
  tmpTail: PdllNode;
begin
  inherited NotifyNodeChange(ANode, Action);
  { correct the list item count }
  case Action of
    nnAdded:
      Inc(FCount);
    nnDeleted:
      Dec(FCount);
    nnCleared:
      FCount := 0;
  end;
  if Action in [nnAdded, nnDeleted] then
  begin
    { if the Node, which is about to destroy, is current cursor then move cursor
      to the next item }
    if Assigned(FCursor) and (ANode = FCursor) then
      FCursor := ANode^.dllnNext;
    { increment the Node Rev ID }
    Inc(FNodeRevID);
    { Notify active mirror lists about the change... }
    if Assigned(FActiveDataUserList) then
    begin
      tmpCursor := FActiveDataUserList.Head^.dllnNext;
      tmpTail := FActiveDataUserList.Tail;
      while tmpCursor <> tmpTail do
      begin
        TgaSharedDoubleList(tmpCursor^.dllnData).NotifyMasterChanged(ANode, Action);
        tmpCursor := tmpCursor^.dllnNext;
      end;
    end;
  end else begin
    { Reset Cursor }
    FCursor := FHead;
    { Clear Node Rev ID }
    FNodeRevID := 0;
    { Notify mirror lists about the clear }
    if Assigned(FDataUserList) then
    begin
      tmpCursor := FDataUserList.Head^.dllnNext;
      tmpTail := FDataUserList.Tail;
      while tmpCursor <> tmpTail do
      begin
        TgaSharedDoubleList(tmpCursor^.dllnData).NotifyMasterChanged(ANode, Action);
        tmpCursor := tmpCursor^.dllnNext;
      end;
    end;
  end;
end;

procedure TgaDoubleList.Pack;
{:
procedure Pack - removes nil items from the list.
}
var
  tmpCursor: PdllNode;
begin
  tmpCursor := Head^.dllnNext;
  while tmpCursor <> FTail do
  begin
    if tmpCursor^.dllnData = nil then
    begin
      tmpCursor := tmpCursor^.dllnNext;
      InternalDelete(tmpCursor^.dllnPrev);
    end else
      tmpCursor := tmpCursor^.dllnNext;
  end;
end;

procedure TgaDoubleList.Previous;
{:
procedure Previous moves the cursor to the previous item in the list.
If cursor is already at the beginning of the list, it deos nothing.
}
begin
  if not Bof then
    FCursor := FCursor^.dllnPrev;
end;

procedure TgaDoubleList.Remove(Item: Pointer);
{:
procedure Remove - removes the first copy of the Item from the list.
}
var
  tmpNode: PdllNode;
begin
  tmpNode := NodeOf(Item);
  if Assigned(tmpNode) then
    InternalDelete(tmpNode);
end;

procedure TgaDoubleList.RemoveActiveDataUser(ADataUser: TgaSharedDoubleList);
{:
procedure RemoveActiveDataUser -  removes ADataUser from the NodeChangeNotify
list. This is called whet the ActiveDataShare := False in the mirror list.
If ADataUser is not foun from the list, this method deos nothing.
}
var
  tmpNode: PdllNode;
begin
  if Assigned(FActiveDataUserList) then
  begin
    tmpNode := FActiveDataUserList.NodeOf(ADataUser);
    if Assigned(tmpNode) then
      FActiveDataUserList.InternalDelete(tmpNode);
  end;
end;

procedure TgaDoubleList.RemoveDataUser(ADataUser: TgaSharedDoubleList);
{:
procedure RemoveDataUser - removes ADataUser from destroy notify list.
Normally called when the mirror list is destroyed.
if ADataUser is not found in list, the method does nothing.
}
var
  tmpNode: PdllNode;
begin
  if Assigned(FDataUserList) then
  begin
    tmpNode := FDataUserList.NodeOf(ADataUser);
    if Assigned(tmpNode) then
      FDataUserList.InternalDelete(tmpNode);
  end;
end;

procedure TgaDoubleList.SetCurrentItem(Value: Pointer);
{:
SetCurrentItem is the write access method of the CurrentItem property.
}
var
  tmpItem: Pointer;
begin
  if Eof or Bof then
    raise EgaLinkListError.Create(SNoCurrentItemForTheOperation);
  tmpItem := Cursor.dllnData;
  if tmpItem <> Value then
  begin
    Cursor.dllnData := Value;
    NotifyChange(tmpItem, Value);
  end;
end;

procedure TgaDoubleList.SetFirstItem(Value: Pointer);
{:
SetFirstItem is the write access method of the FirstItem property.
It sets the first item (the one after the Head node) in the list.
If the list is empty, the exeception is raised.
}
var
  tmpItem: Pointer;
begin
  if IsEmpty then
    raise EgaLinkListError.Create(SNoItemsInTheList);
  tmpItem := Head^.dllnNext.dllnData;
  if tmpItem <> Value then
  begin
    Head^.dllnNext.dllnData := Value;
    NotifyChange(tmpItem, Value);
  end;
end;

procedure TgaDoubleList.SetLastItem(Value: Pointer);
{:
SetLastItem is the write access method of the LastItem property.
It sets the last item (the one before the Tail node) in the list.
If the list is empty, the exeception is raised.
}
var
  tmpItem: Pointer;
begin
  if IsEmpty then
    raise EgaLinkListError.Create(SNoItemsInTheList);
  tmpItem := Tail^.dllnPrev.dllnData;
  if tmpItem <> Value then
  begin
    Tail^.dllnPrev.dllnData := Value;
    NotifyChange(tmpItem, Value);
  end;
end;

{
***************************** TgaSharedDoubleList ******************************
}
constructor TgaSharedDoubleList.Create;
{:
Constructor Create overrides the inherited Create.
First the internal data structure is initialized to indicate that the list
itself is the owner of the data, then inherited Create is
called.
}
begin
  FDataOwner := Self;
  FIsDataOwner := True;
  inherited Create;
end;

constructor TgaSharedDoubleList.CreateMirror(AMirroredList: TgaDoubleList);
{:
constructor CreateMirror - creates a "mirror" list for the AMirroredList. The
lists will have common dataowner (i.e. if Ythe AMirroredList is itself a mirror,
then the dataowner will be the dataowner of the AMirroredList). Note that the
DataIwner of TgaDoubleList is implicitly the list itself.
The contents of the list will be exactly the same as it is for AMirroredList -
if it has strict last or first item, a new mirror will have the same.
the starting/ening postions can be changed later trough the call to
SetStartPos/SetEndPos.
}
var
  tmpActiveMirror: TgaSharedDoubleList;
begin
  FHead := AMirroredList.Head;
  FTail := AMirroredList.Tail;
  if AMirroredList is TgaSharedDoubleList then
  begin
    tmpActiveMirror := TgaSharedDoubleList(AMirroredList);
    FDataOwner := tmpActiveMirror.DataOwner;
    FFirstNode := tmpActiveMirror.FFirstNode;
    FLastNode := tmpActiveMirror.FLastNode;
  end else
    FDataOwner := AMirroredList;
  DataOwner.AddDataUser(Self);
  DataOwner.AddActiveDataUser(Self);
  FActiveDataShare := True;
  FIsDataOwner := False;
  FNodeRevID := AMirroredList.NodeRevID;
  FCursorNodeRev := DataOwner.NodeRevID;
  inherited Create;
end;

destructor TgaSharedDoubleList.Destroy;
{:
Destructor Destroy overrides the inherited Destroy.
If the list is a mirroring list, then the list must be removed from the
(Active)DataUserList of the DataOwner. Also, the Head and Tail must be cleared
to avoid changing the master list.
Finally inherited Destroy is called.
}
begin
  if not IsDataOwner then
  begin
    if Assigned(FDataOwner) then
    begin
      ActiveDataShare := False;
      FDataOwner.RemoveDataUser(Self);
    end;
    FTail := nil;
    FHead := nil;
    FFirstNode := nil;
    FLastNode := nil;
  end;
  inherited Destroy;
end;

procedure TgaSharedDoubleList.Add(Item: Pointer);
{:
procedure Add overrides inherited Add.
This method is overridden to allow the strict mirror list to grow, if the
Add method is called on this instance. To do that, the FirstNode and LastNode
are unbound, then inherited Add is called. After that, theese nodes are rebound,
if needed.
}
var
  WasStrictLast: Boolean;
  FirstNodeWasUnbound: Boolean;
begin
  WasStrictLast := Assigned(FLastNode);
  FLastNode := nil;
  // Unbind the first node, if list is strictempty
  FirstNodeWasUnbound := FFirstNode = FTail;
  if FirstNodeWasUnbound then
    FFirstNode := nil;
  inherited Add(Item);
  if WasStrictLast then
    FLastNode := Tail^.dllnPrev;
  if FirstNodeWasUnbound then
    FFirstNode := Head^.dllnNext;
end;

procedure TgaSharedDoubleList.AddActiveDataUser(ADataUser: TgaSharedDoubleList);
{:
procedure AddActiveDataUser overrides inherited AddActiveDataUser.
If the list is not a Data Owner, then an exception is raised to indicate that
mirroring lists can't have data users.
}
begin
  if not IsDataOwner then
    raise EgaLinkListError.CreateFmt(SListIsNotDataOwner, [SCantAddActiveDataUser]);
  inherited AddActiveDataUser(ADataUser);
end;

procedure TgaSharedDoubleList.AddDataUser(ADataUser: TgaSharedDoubleList);
{:
procedure AddDataUser overrides inherited AddDataUser calls
If the list is not a Data Owner, then an exception is raised to indicate that
mirroring lists can't have data users.
}
begin
  if not IsDataOwner then
    raise EgaLinkListError.CreateFmt(SListIsNotDataOwner, [SCantAddDataUser]);
  inherited AddDataUser(ADataUser);
end;

procedure TgaSharedDoubleList.CheckListBorders;
{:
procedure CheckListBorders overrides inherited donothing
TgaSimpleDoubleList.CheckListBorders. For optimization, the call to inherited
CheckListBorders is omitted.
ValidateListBorders is called if required - the list is mirroring list and it
is not in the DataOwner's ActiveDataUser list (ActiveDataShare = False)
ValidateListBorders corrects Head and Tail nodes, if needed and possible.
If unrepairable error occures, an exception is raised.
}
begin
  // call to inherited do-nothing method not required??
  if not IsDataOwner then
    if not ActiveDataShare then
      ValidateListBorders;
end;

procedure TgaSharedDoubleList.CheckListCursor;
{:
procedure CheckListCursor overrides inherited donothing
TgaSimpleDoubleList.CheckListCursor. For optimization, the call to inherited
CheckListCursor is omitted.
If cursor is not inside the list, and exception is raised in ValidateCursor
method.
}
begin
  // call to inherited do-nothing method not required??
  if not IsDataOwner then
    if (not ActiveDataShare) then
      ValidateCursor(False);
end;

procedure TgaSharedDoubleList.Clear;
{:
procedure Clear overrides inherited Clear.
The inherited Clear is called only if the list is a owner of the data.
If not, the list is emptied item by item.
}
var
  Temp: PdllNode;
  tmpData: Pointer;
  FFirstNodeWasBound, FLastNodeWasBound: Boolean;
begin
  if IsDataOwner then
    inherited Clear
  else begin
    if Assigned(FHead) then
    begin
      Temp := Head^.dllnNext;
      FFirstNodeWasBound := Assigned(FFirstNode);
      FFirstNode := nil;
      FLastNodeWasBound := Assigned(FLastNode);
      FLastNode := nil;
      while (Temp <> FTail) do begin
        FHead^.dllnNext := Temp^.dllnNext;
        NotifyNodeChange(Temp, nnDeleted);
        tmpData := Temp^.dllnData;
        dnmFreeNode(Temp);
        if Assigned(tmpData) then
          Notify(tmpData, lnDeleted);
        Temp := FHead^.dllnNext;
      end;
      FHead^.dllnNext := FTail;
      FTail^.dllnPrev := FHead;
      { #Todo2 There will be conflict if the list has strict starting position
       but floating end position - should it be kept empty or not }
      if FFirstNodeWasBound  then
        FFirstNode := FTail;
      if FLastNodeWasBound then
        FLastNode := FHead;
    end;
    // FCount := 0; Do not the count as for mirror list the count is always -1
    // if the list had strict first or last node, then correct them so that the
    // list would not start growing
  end;
end;

procedure TgaSharedDoubleList.DataOwnerDestroyed;
{:
procedure DataOwnerDestroyed is called when the DataOwner list is destroyed.
As all list data is also destroyed, the list is marked as "invalid"
}
begin
  FDataOwner := nil;
  FHead := nil;
  FTail := nil;
  FFirstNode := nil;
  FLastNode := nil;
  FCursor := nil;
end;

procedure TgaSharedDoubleList.InitDoubleList;
{:
procedure InitDoubleList overrides inherited InitDoubleList.
If the list is DataOwner, the inherited InitDoubleList is called.
If it is not, the Cursor it inited to the Head node and the Count is set to -1 -
tom indicate that the count is not valid for the mirror list
(all other initialization is done in CreateMirror constructor)
}
begin
  if IsDataOwner then
    inherited InitDoubleList
  else begin
    {set the cursor to the head node}
    FCursor := FHead;
    FCount := -1;
  end;
end;

procedure TgaSharedDoubleList.InsertAfterCurrent(AItem: Pointer; 
        MoveCursorToNewItem: boolean);
{:
procedure InsertAfterCurrent overrides inherited InsertAfterCurrent.
Reason: to allow the mirroring list with strict start or end position to
"absorb" a item if the current item is the last one in the list
}
var
  FirstNodeWasUnbound: Boolean;
begin
  // Unbind the first node
  FirstNodeWasUnbound := FFirstNode = FTail;
  if FirstNodeWasUnbound then
    FFirstNode := nil;
  if FLastNode <> Cursor then
    inherited InsertAfterCurrent(AItem, MoveCursorToNewItem)
  else begin
    FLastNode := InternalInsert(AItem, Cursor);
    FTail := FLastNode^.dllnNext;
    if MoveCursorToNewItem then
      FCursor := FLastNode;
  end;
  // rebind first node if it was bound before
  if FirstNodeWasUnbound then
    FFirstNode := Head^.dllnNext;
end;

procedure TgaSharedDoubleList.Notify(Ptr: Pointer; Action: TListNotification);
{:
procedure Notify overrides inherited Notify calls inherited Notify.
Also calls the dataowners Notify, if the lis is mirror  list
}
begin
  inherited Notify(Ptr, Action);
  if not IsDataOwner then
    DataOwner.Notify(Ptr, Action);
end;

procedure TgaSharedDoubleList.NotifyMasterChanged(ANode: PdllNode; Action: 
        TgaNodeNotification);
{:
procedure NotifyMasterChanged.
Occurs for mirroring lists if the list is in ActiveDataShare state,
This method should be dispatched after the NotifyNodeChange is finished
ANode - a node which is changing, Action - action to be performed.
}
begin
  case Action of
    nnAdded: begin
      if ANode^.dllnNext = FFirstNode then
      begin
        FHead := ANode;
        if ANode^.dllnPrev = FLastNode then
          FLastNode := ANode;
      end;
      if ANode^.dllnPrev = FLastNode then
        FTail := ANode;
    end;
    nnDeleted: begin
      if ANode = FCursor then
        if Eof then
          Last
        else
          Next;
      if ANode = FTail then
        if Assigned(FLastNode) then
          FTail := FLastNode^.dllnNext
        else
          FTail := ANode^.dllnNext;
      if ANode = FHead then
        if Assigned(FFirstNode) then
          FHead := FFirstNode^.dllnPrev
        else
          FHead := ANode^.dllnPrev;
      if ANode = FLastNode then
        if FFirstNode = Tail then
          FlastNode := Head
        else
          FLastNode := FLastNode^.dllnPrev;
      if ANode = FFirstNode then
        if FLastNode = Head then
          FFirstNode := Tail
        else
          FFirstNode := FFirstNode^.dllnNext;
    end;
    nnCleared: begin
      FHead := DataOwner.Head;
      FTail := DataOwner.Tail;
      if Assigned(FFirstNode) then
        FFirstNode := FTail;
      if Assigned(FLastNode) then
        FLastNode := FHead;
      FCursor := Head;
    end;
  end;
  FNodeRevID := DataOwner.NodeRevID;
  FCursorNodeRev := DataOwner.NodeRevID;
end;

procedure TgaSharedDoubleList.NotifyNodeChange(ANode: PdllNode; Action: 
        TgaNodeNotification);
{:
procedure NotifyNodeChange overrides inherited NotifyNodeChange.
If list IsDataOwner, the calls inherited NotifyNodeChange, otherwise calls
DataOwners NotifyNodeChange
}
begin
  if IsDataOwner then
    inherited NotifyNodeChange(ANode, Action)
  else
    DataOwner.NotifyNodeChange(ANode, Action);
end;

procedure TgaSharedDoubleList.RemoveActiveDataUser(ADataUser: 
        TgaSharedDoubleList);
{:
procedure RemoveActiveDataUser overrides inherited RemoveActiveDataUser.
Calls inherited RemoveActiveDataUser if the list is data owner, otherwise
raises an exception.
}
begin
  if not IsDataOwner then
    raise EgaLinkListError.CreateFmt(SListIsNotDataOwner, [SCantRemoveActiveDataUser]);
  inherited RemoveActiveDataUser(ADataUser);
end;

procedure TgaSharedDoubleList.RemoveDataUser(ADataUser: TgaSharedDoubleList);
{:
procedure RemoveDataUser overrides inherited RemoveDataUser.
Calls inherited RemoveDataUser if the list is data owner, otherwise
raises an exception.
}
begin
  if not IsDataOwner then
    raise EgaLinkListError.CreateFmt(SListIsNotDataOwner, [SCantRemoveDataUser]);
  inherited RemoveDataUser(ADataUser);
end;

procedure TgaSharedDoubleList.SetActiveDataShare(Value: Boolean);
{:
SetActiveDataShare is the write access method of the ActiveDataShare property.
The ACtiveDataShare can be changed only if the list is a mirror list.
If ActiveDataShare is disabled, the list is removed from
DataOwners.ActiveDataUsers list.
If ActiveDataShare is enabled, the list borders and cursor are validated, and
the list is added to the DataOwners ActiveDataUsers list.
}
begin
  if FActiveDataShare <> Value then
  begin
    if IsDataOwner then
      raise EgaLinkListError.CreateFmt(SListIsDataOwner, [SCantChangeActiveDataShare]);
    if Value then
    begin
      ValidateListBorders;
      ValidateCursor(True);
      DataOwner.AddActiveDataUser(Self);
    end else
      DataOwner.RemoveActiveDataUser(Self);
    FActiveDataShare := Value;
  end;
end;

procedure TgaSharedDoubleList.SetEndPos(APositionedList: TgaDoubleList; 
        StrictEndPos: Boolean);
{:
procedure SetEndPos - Sets the lits's last item to be the current item of the
APostionedList. If StrictEnPos is true, the items inserted to the DataOwnerList
between the last item and the item after that, are not accompanied to this list.
If StrictEnPos is False, all items added before the item next to the last item
set are absorbed into the list.
If the PostionedList is at Eof, then StrictEndPos do not have any effect.
APositionedList can be self.
}
begin
  if IsDataOwner then
    raise EgaLinkListError.CreateFmt(SListIsNotAMirroringList, [SCantSetTheStartOrEndPostions]);
  if (DataOwner <> APositionedList) and
    ((APositionedList is TgaSharedDoubleList) and
      (DataOwner <> (APositionedList as TgaSharedDoubleList).DataOwner)) then
    raise EgaLinkListError.CreateFmt(SListsDoNotShareData, [SCantSetTheStartOrEndPostions]);
  if APositionedList.Eof then
  begin
    FTail := APositionedList.Cursor;
    FLastNode := nil;
  end else
  begin
    FTail := APositionedList.FCursor^.dllnNext;
    if StrictEndPos then
      FLastNode := FTail^.dllnPrev
    else
      FLastNode := nil;
  end;
  // the node rev id-s must be reste before the validate
  FNodeRevID := 0;
  FCursorNodeRev := 0;
  ValidateListBorders;
  ValidateCursor(True);
end;

procedure TgaSharedDoubleList.SetStartPos(APositionedList: TgaDoubleList; 
        StrictStartPos: boolean);
{:
procedure SetStartPos - Sets the lits's first item to be the current item of the
APostionedList. If StrictStartPos is true, the items inserted to the
DataOwnerList between the first item and the item before that, are not
accompanied to this list. If StrictEnPos is False, all items added after the
item previous to the last item set are absorbed into the list.
If the PostionedList is at Bof, then StrictStartPos do not have any effect.
APositionedList can be self.
}
begin
  if IsDataOwner then
    raise EgaLinkListError.CreateFmt(SListIsNotAMirroringList, [SCantSetTheStartOrEndPostions]);
  if (DataOwner <> APositionedList) and
    ((APositionedList is TgaSharedDoubleList) and
      (DataOwner <> (APositionedList as TgaSharedDoubleList).DataOwner)) then
    raise EgaLinkListError.CreateFmt(SListsDoNotShareData, [SCantSetTheStartOrEndPostions]);
  if APositionedList.Bof then
  begin
    FHead := APositionedList.Cursor;
    FFirstNode := nil;
  end else
  begin
    FHead := APositionedList.Cursor^.dllnPrev;
    if StrictStartPos then
      FFirstNode := FHead^.dllnNext
    else
      FFirstNode := nil;
  end;
  // the node rev id-s must be reste before the validate
  FNodeRevID := 0;
  FCursorNodeRev := 0;
  ValidateListBorders;
  ValidateCursor(True);
end;

procedure TgaSharedDoubleList.ValidateCursor(CorrectIfInvalid: Boolean);
{:
procedure ValidateCursor - Checks the cursor of the list. If it is postioned
incorrectly, the if CorrectIfInvalid = True, Cursor is set to be at the Bof.
If False, an exception is raised.
}
begin
  if FCursorNodeRev <> DataOwner.NodeRevID then
  begin
    if not IsCursorCorrect then
    begin
      if CorrectIfInvalid then
        FCursor := Head
      else
        raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, [SInvalidCursor]);
    end;
    FCursorNodeRev := DataOwner.NodeRevID;
  end;
end;

procedure TgaSharedDoubleList.ValidateListBorders;
{:
procedure ValidateListBorders - checks if the list Tail and Head are inside the
dataowners list, and that the Strict astrt and end postions are correctly
preserved.
Raises an exception if the list can't be validated
}
begin
  if not IsDataOwner then
  begin
    { if the Master list node Revision ID is not changed, exit - list is up to
    date }
    if NodeRevID = DataOwner.NodeRevID then
      Exit;
    { Check wheter FFirstItem and FLastItem are part of master list. if are,
      then Tail and Head can be silently set (As they are used only for
      convience)}
    if Assigned(FFirstNode) then
    begin
      if DataOwner.IsNodeInside(FFirstNode) then
      begin
        if FFirstNode = DataOwner.Head then
          raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, [SStrictFirstNodePointsToTheHeadOfMasterList]);
        FHead := FFirstNode^.dllnPrev;
        if (FFirstNode = FTail) and Assigned(FLastNode) then
          FLastNode := FHead;
      end else
        raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, [SStrictFirstNodeNotPartOfTheMasterList]);
    end else
      if not DataOwner.IsNodeInside(FHead) then
        raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, [SHeadNodeNotInMasterList]);
    if Assigned(FLastNode) then
      FTail := FLastNode^.dllnNext;
    { begin
       it is not strictly required to check wheter the tail node is inside the
        master list. As the head node is (verified), then if
        IsList(Head, Tail) = True, and provided that the master list is not
        corrupted, the last node has to be inside master list
      if DataOwner.IsNodeInside(FLastNode) then
      begin
        if FLastNode = DataOwner.Tail then
          raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, ['Strict last node is invalid - points to the Tail item of master list']);
        FTail := FLastNode^.dllnNext;
       end else
        raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, ['Strict last node is invalid - not part of the master list']);
    end else
      if not DataOwner.IsNodeInside(FTail) then
        raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, ['Tail node is invalid - not part of the master list']);}
    if not IsList(FHead, FTail) then
      raise EgaLinkListError.CreateFmt(SMirrorListValidateFailed, [SInvalidHeadOrTail]);
    { remember the last validate time }
    FNodeRevID := DataOwner.NodeRevID;
  end;
end;

{
**************************** TgaDoubleListBookmark *****************************
}
constructor TgaDoubleListBookmark.Create(AOwnerList: TgaDoubleList);
{:
Constructor Create overrides the inherited Create. 
First inherited Create is called, then the internal data structure is
initialized to store tha list for which a bookmark was created and current
postion inside this list. 
}
begin
  inherited Create;
  FOwnerList := AOwnerList;
  FCursor := FOwnerList.Cursor;
  FNodeRevID := FOwnerList.NodeRevID;
end;

function TgaDoubleListBookmark.GetIsValid: Boolean;
{:
function GetIsValid overrides inherited GetIsValid.
Returns: True, if the Bookmark is valid
False, if it is not.
}
begin
  if NodeRevID = OwnerList.NodeRevID then
    Result := True
  else begin
    Result := OwnerList.IsNodeInside(Cursor);
    if Result then
      FNodeRevID := OwnerList.NodeRevID;
  end;
end;

function TgaDoubleListBookmark.GetItem: Pointer;
{:
function GetItem overrides inherited GetItem.
Returns: A pointer to the item represented by the Bookmark
}
begin
  if not IsValid then
    raise EgaListBookmarkError.CreateFmt(SInvalidBookmark, [SCantGetItem]);
  Result := Cursor.dllnData;
end;

function TgaDoubleListBookmark.IsValidFor(AList: TgaDoubleList): Boolean;
{:
function IsValidFor checks if the bookmark is valid for a list represented by
the AList.
Returns: True, if the AList uses the same data as bookmarks owner list
(AList = OwnerList or AList.DataOwner = OwnerList.DataOwner, if this compare
is (partially) applicable), and the node represented by the bookmark's cursor is
inside the AList
}
begin
  // use IsValid if possible as IsValid is more optimized
  if AList = OwnerList then
    Result := IsValid
  else begin
    if (AList is TgaSharedDoubleList) and (OwnerList is TgaSharedDoubleList) then
      Result := TgaSharedDoubleList(OwnerList).DataOwner = TgaSharedDoubleList(AList).DataOwner
    else if AList is TgaSharedDoubleList then
      Result := OwnerList = TgaSharedDoubleList(AList).DataOwner
    else if OwnerList is TgaSharedDoubleList then
      Result := TgaSharedDoubleList(OwnerList).DataOwner = AList
    else
      Result := False;
    if not Result then
      Exit;
    Result := AList.IsNodeInside(Cursor);
  end;
end;

procedure TgaDoubleListBookmark.SetItem(Value: Pointer);
{:
procedure SetItem overrides inherited SetItem.
Changes the item represented by the bookmark to the new Value.
}
var
  tmpItem: Pointer;
begin
  if not IsValid then
    raise EgaListBookmarkError.CreateFmt(SInvalidBookmark, [SCantSetItem]);
  tmpItem := Cursor^.dllnData;
  if tmpItem <> Value then
  begin
    Cursor^.dllnData := Value;
    OwnerList.NotifyChange(tmpItem, Value);
  end;
end;

end.
